import mock
import random
import shlex
import unittest
from bson.objectid import ObjectId

from kaiju.cli import KaijuCLI

from . import little_app


class KaijuTests(unittest.TestCase):

    def setUp(self):
        self.cli = KaijuCLI(little_app.trilotest.operations,
                            description="Little test app")
        little_app.trilotest.drop_database()

    def tearDown(self):
        little_app.trilotest.reset_flags()
        little_app.trilotest.drop_database()

    def test_generation(self):
        s_a, s_b = -5, 6
        s_size = 20
        s_seed = 1138
        a = little_app.generate_uniform(s_a, s_b, sample_size=s_size,
                                        seed=s_seed)

        random.seed(s_seed)
        reference_data = [random.randint(s_a, s_b) for _ in xrange(s_size)]
        self.assertIsInstance(a, ObjectId)

        recorded_data = little_app.trilotest.fetch(a, collection="bigdata")
        self.assertItemsEqual(recorded_data, reference_data)

    def test_run_sequence(self):
        a = little_app.generate_uniform(-5, 5, 100, seed=1138)
        b = little_app.squared(sample=a)

        stored_log = little_app.trilotest.fetch_log(a)
        stored_log.pop("date")
        self.assertEqual(
            stored_log,
            {
                "_id": a,
                "operation": "generate_uniform",
                "kwargs": {"a": -5, "b": 5, "sample_size": 100, "seed": 1138},
                "output": "bigdata",
            }
        )

        stored_log = little_app.trilotest.fetch_log(b)
        stored_log.pop("date")
        self.assertEqual(
            stored_log,
            {
                u"_id": b,
                u"operation": u"squared",
                u"kwargs": {u"sample": a},
                u"output": u"bigdata",
            }
        )

        self.assertAlmostEqual(little_app.mean(sample=a).next(), -0.31,
                               delta=1e-12)
        self.assertAlmostEqual(little_app.mean(sample=b).next(), 9.45,
                               delta=1e-12)

    def test_names(self):
        self.assertItemsEqual(
            little_app.trilotest.operations.keys(),
            ["generate_uniform", "squared", "mean", "some_arguments_testing",
             "provide_pokemons"]
        )

    def test_var(self):
        test_var_name = "__TRILOBITE_TEST_VARIABLE"

        self.assertIsNone(little_app.trilotest.var_get(test_var_name).next())

        # Set variable
        test_value_1 = 12321
        result_1 = little_app.trilotest.var_set(test_var_name,
                                                test_value_1).next()

        self.assertEqual(result_1["ok"], 1.0)
        self.assertEqual(result_1["n"], 1)
        self.assertFalse(result_1["updatedExisting"])
        self.assertIsInstance(result_1["upserted"], ObjectId)

        # Get variable value
        self.assertEqual(little_app.trilotest.var_get(test_var_name).next(),
                         test_value_1)

        # Update value
        test_value_2 = 65456
        result_2 = little_app.trilotest.var_set(test_var_name,
                                                test_value_2).next()

        self.assertEqual(result_2["ok"], 1.0)
        self.assertEqual(result_2["n"], 1)
        self.assertTrue(result_2["updatedExisting"])

        # List variables
        self.assertEqual(little_app.trilotest.var_get(test_var_name).next(),
                         test_value_2)

        result_ls = little_app.trilotest.var_ls().next()
        self.assertEqual(result_ls["name"], test_var_name)
        self.assertEqual(result_ls["value"], test_value_2)

        # Delete variable
        result_del = little_app.trilotest.var_del(test_var_name).next()
        result_del["ok"] == 1.0
        result_del["n"] == 1

        # Check if it was really deleted
        self.assertEqual(list(little_app.trilotest.var_ls()), [])

    def _execute_shell(self, shell_cmd):
        gargs, args = self.cli.parse_args(shlex.split(shell_cmd))
        little_app.trilotest.handle_kaiju_arguments(gargs)
        operation = gargs["operation"]
        return little_app.trilotest.execute(operation, args)

    def test_simple_usage_and_defaults(self):
        [output] = list(self._execute_shell("some_arguments_testing 1 0.2"))
        self.assertEqual(output, dict(my_bool=False, my_int=1, my_float=0.2,
                                      my_string="string"))

    def test_bool_store_true(self):
        [output] = list(self._execute_shell(
            "some_arguments_testing 1 0.2 --my_bool"
        ))
        self.assertEqual(output, dict(my_bool=True, my_int=1, my_float=0.2,
                                      my_string="string"))

    def test_arguments_default_parse(self):
        [output] = list(self._execute_shell(
            "some_arguments_testing 1 0.2 --my_string str"
        ))
        self.assertEqual(output, dict(my_bool=False, my_int=1, my_float=0.2,
                                      my_string="str"))

    @mock.patch("sys.stderr", mock.Mock())
    def test_type_check_int_can_be_float(self):
        with self.assertRaises(SystemExit):
            self._execute_shell("some_arguments_testing 0.1 2")

    @mock.patch("sys.stderr", mock.Mock())
    def test_type_check_int_cant_be_str(self):
        with self.assertRaises(SystemExit):
            self._execute_shell("some_arguments_testing abc 2")

    @mock.patch("sys.stderr", mock.Mock())
    def test_type_check_float_cant_be_str(self):
        with self.assertRaises(SystemExit):
            self._execute_shell("some_arguments_testing 1 abc")

    @mock.patch("sys.stderr", mock.Mock())
    def test_mandatory_arguments_are_mandatory(self):
        with self.assertRaises(SystemExit):
            self._execute_shell("some_arguments_testing")

    def test_data_argument_parse(self):
        oid1 = ObjectId("a" * 24)
        oid2 = ObjectId("b" * 24)
        little_app.trilotest.database.pokemon.insert([
            {"maker_id": oid1, "value": {"type": "fire", "name": "Charizard"}},
            {"maker_id": oid1, "value": {
                "type": "water", "name": "Blastoise"}},
            {"maker_id": oid1, "value": {"type": "grass", "name": "Venusaur"}},
            {"maker_id": oid2, "value": {
                "type": "fire", "name": "Typhlosion"}},
            {"maker_id": oid2, "value": {"type": "water", "name": "Feraligatr"}},
            {"maker_id": oid2, "value": {"type": "grass", "name": "Meganium"}},
        ])

        cmd = r"provide_pokemons {}".format("a" * 24)
        pokemons = list(self._execute_shell(cmd))

        self.assertItemsEqual(
            pokemons,
            [
                {"type": "fire", "name": "Charizard"},
                {"type": "water", "name": "Blastoise"},
                {"type": "grass", "name": "Venusaur"},
            ]
        )

    def test_memoized_results_arent_recalculated(self):
        cmd = "generate_uniform 0 10 10"
        first_run = self._execute_shell(cmd)
        second_run = self._execute_shell(cmd)

        self.assertEqual(first_run, second_run)

    def test_no_memo_forces_execution(self):
        cmd = "generate_uniform 0 10 10"
        cmd_no_memo = "--no-memo generate_uniform 0 10 10"

        results = []
        results.append(self._execute_shell(cmd))
        results.append(self._execute_shell(cmd))
        results.append(self._execute_shell(cmd_no_memo))

        self.assertEqual(results[0], results[1])
        self.assertNotEqual(results[1], results[2])

    def test_no_store_flag(self):
        cmd = "generate_uniform 0 10 10 --seed 1138"
        result1 = self._execute_shell(cmd)
        cmd = "--no-store generate_uniform 0 10 10 --seed 1138"
        result2 = self._execute_shell(cmd)

        self.assertEqual(type(result1), ObjectId)
        self.assertEqual(
            list(result2),
            [3, 5, 3, 1, 7, 4, 3, 8, 4, 1]
        )



"""
Kaiju is a framework created out of necessity to organize and build a
common interface to lots of standalone scripts and different kinds of data
analysis.

It saves a log of the operations used into MongoDB, as explained in the README.
There's also a command line interface which exposes all the operations defined.

"""

from bson.errors import InvalidDocument
from datetime import datetime
from functools import wraps
from inspect import getargspec
from pymongo import MongoClient
import sys


class KaijuException(Exception):
    pass


class FakeDatabase(object):
    def __getitem__(self, _):
        raise KaijuException("Kaiju application not connected to database.")


class Kaiju(object):
    LOG_COLLECTION = "kaiju_log"
    VAR_COLLECTION = "kaiju_var"

    def __init__(self, database_name):
        self.database_name = database_name
        self.database = FakeDatabase()

        self.operations = {}

        self.reset_flags()

    def reset_flags(self):
        self.allow_memoize = True
        self.allow_store = True

    def handle_kaiju_arguments(self, gargs):
        self.connect_database(gargs["database_uri"])

        self.reset_flags()
        self.enable_print = gargs["enable_print"]
        if gargs["no_memo"]:
            self.set_no_memo()
        if gargs["no_store"]:
            self.set_no_store()

    def connect_database(self, database_uri=None):
        self.mongo_client = MongoClient(database_uri)
        self.database = self.mongo_client[self.database_name]

        self.database[self.VAR_COLLECTION].ensure_index("name", 1)
        self.database[self.LOG_COLLECTION].ensure_index("date", 1)
        self.database[self.LOG_COLLECTION].ensure_index("operation", 1)

        for operation in self.operations.values():
            kaiju_parameters = operation.__kaiju__
            if "output" in kaiju_parameters:
                self.database[kaiju_parameters["output"]].ensure_index("maker_id", 1)

    def set_no_memo(self):
        self.allow_memoize = False

    def set_no_store(self):
        self.allow_store = False

    def __call__(self, **kaiju_parameters):
        def decorator(operation):
            @wraps(operation)
            def wrapper(*args, **kwargs):
                dictified_input_args = self.get_dictified_input_args(
                    operation, kaiju_parameters, args, kwargs
                )

                ## If operation has been executed previously, return
                ## stored result. Proceed with execution otherwise.
                if self.allow_memoize and self.allow_store:
                    existing_log_entry = self.log_query(operation.__name__, dictified_input_args)
                    if existing_log_entry is not None:
                        return existing_log_entry["_id"]

                if self.enable_print:
                    self.stdout_backup = sys.stdout
                    sys.stdout = sys.stderr

                new_kwargs = self.process_operation_args(
                    operation, kaiju_parameters, args, kwargs
                )

                result = operation(**new_kwargs)

                if self.enable_print:
                    sys.stdout = self.stdout_backup

                if "output" in kaiju_parameters and self.allow_store:
                    maker_id = self.log_add(
                        operation.__name__,
                        dictified_input_args,
                        kaiju_parameters["output"]
                    )
                    try:
                        self.store(kaiju_parameters["output"], result, maker_id)
                    except Exception:
                        self.delete(maker_id)
                        raise
                    return maker_id
                else:
                    return result

            wrapper.__original__ = operation
            wrapper.__kaiju__ = kaiju_parameters

            self.operations[operation.__name__] = wrapper
            return wrapper
        return decorator

    def execute(self, operation_name, kwargs):
        return self.operations[operation_name](**kwargs)

    def fetch(self, maker_id, collection=None, mongo_cursor=False):
        """Get documents from `collection` according to `maker_id`."""

        if collection is None:
            collection = self.database[self.LOG_COLLECTION].find_one(
                {"_id": maker_id}
            )["output"]

        query = self.database[collection].find({"maker_id": maker_id})
        if not mongo_cursor:
            return (x["value"] for x in query)
        else:
            return query

    def count(self, maker_id):
        """Count documents from `collection` according to `maker_id`."""

        return self.fetch(maker_id, mongo_cursor=True).count()

    def log_add(self, operation_name, kwargs, output_collection):
        return self.database[self.LOG_COLLECTION].insert(
            {
                "operation": operation_name,
                "kwargs": kwargs,
                "date": datetime.utcnow(),
                "output": output_collection
            }
        )

    def log_query(self, operation_name, kwargs):
        query = self.database[self.LOG_COLLECTION].find(
            {
                "operation": operation_name,
                "kwargs": kwargs
            }
        ).sort("date", -1).limit(1)
        query = list(query)
        return query[0] if query else None

    def log_find(self, *args, **kwargs):
        return self.database[self.LOG_COLLECTION].find(
            *args, **kwargs
        )

    def store(self, output_collection, data, maker_id):
        self.database[output_collection].insert(
            {
                "maker_id": maker_id,
                "value": x
            }
            for x in data
        )

    def delete(self, maker_id):
        run = self.database[self.LOG_COLLECTION].find_one({"_id": maker_id})
        if run is None:
            return False

        output_collection = run["output"]

        del_data = self.database[output_collection].remove({"maker_id": maker_id})["err"]
        del_log = self.database[self.LOG_COLLECTION].remove({"_id": maker_id})["err"]
        return (del_data is None and del_log is None)

    def fetch_log(self, operation_id=None):
        if operation_id is None:
            return self.database[self.LOG_COLLECTION].find().sort([("date", -1)]).limit(1000)
        else:
            return self.database[self.LOG_COLLECTION].find_one({"_id": operation_id})

    def drop_database(self):
        self.mongo_client.drop_database(self.database_name)

    def get_dictified_input_args(self, operation, kaiju_parameters, args, kwargs):
        operation_args = getargspec(operation)[0]

        dictified_input_args = {}
        dictified_input_args.update(kwargs)
        dictified_input_args.update(zip(operation_args, args))
        return dictified_input_args

    def process_operation_args(self, operation, kaiju_parameters, args, kwargs):
        operation_args = getargspec(operation)[0]

        new_kwargs = {
            k: (kaiju_parameters[k].process(self, v)
                if k in kaiju_parameters else v)
            for k, v in kwargs.items()
        }
        new_kwargs.update(
            {
                k: (kaiju_parameters[k].process(self, v)
                    if k in kaiju_parameters else v)
                for k, v in zip(operation_args, args)
            }
        )
        return new_kwargs

    def var_set(self, name, value):
        "Assign `value` to persitent variable `name`."

        if value is None:
            self.var_del(name)

        yield self.database[self.VAR_COLLECTION].update(
            {"name": name},
            {"name": name, "value": value},
            upsert=True
        )

    def var_del(self, name):
        yield self.database[self.VAR_COLLECTION].remove({"name": name})

    def var_get(self, name):
        query = self.database[self.VAR_COLLECTION].find_one({"name": name})
        yield query["value"] if query is not None else None

    def var_ls(self):
        return ({"name": x["name"], "value": x["value"]}
                for x in self.database[self.VAR_COLLECTION].find())



import argparse
import os
from inspect import getargspec

from kaiju.arguments import Var


class KaijuCLI(object):
    """KaijuCLI simplifies the creation of a command line interface.

    When instantiating a new KaijuCLI, youa provide a list of operations and a
    description. A subparser is automatically created for each operation,
    giving you a basic interface (including a useful --help option) with no
    cost.
    """

    SPECIAL_ARG_NAMES = ["operation", "database_uri", "no_memo", "no_store", "enable_print"]

    def __init__(self, operations, description):
        self.operations = operations

        self.parser = argparse.ArgumentParser(
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        self.parser.add_argument(
            "--database-uri",
            default=os.environ.get("TRILOBITE_DATABASE_URI", None)
        )
        self.parser.add_argument("--no-memo", default=False, action="store_true")
        self.parser.add_argument("--no-store", default=False, action="store_true")
        self.parser.add_argument("--enable-print", default=False, action="store_true")

        self.subparsers = self.parser.add_subparsers(title="operations", dest="operation")

        for op_name, op in self.operations.items():
            self.configure_parser_from_function(op)

    def configure_parser_from_function(self, op):
        """Map arguments from `op` to parser entries.

        Analyze args from `op` and create corresponding entries on the parser.
        """

        brief_description = op.__doc__.splitlines() if op.__doc__ else [""]
        sp = self.subparsers.add_parser(op.__name__, help=brief_description[0])
        args, _, _, defaults = getargspec(op.__original__)
        if defaults is None:
            defaults = []

        cut = len(args) - len(defaults)
        non_optional_arguments = args[:cut]
        optional_arguments = args[cut:]

        kaiju_arguments = op.__kaiju__

        for argument in non_optional_arguments:
            arg_handler = kaiju_arguments.get(argument, Var())
            sp.add_argument(argument, **arg_handler.config)

        for argument, default_value in zip(optional_arguments, defaults):
            arg_handler = kaiju_arguments.get(argument, Var())
            sp.add_argument("--" + argument,
                            default=default_value,
                            **arg_handler.config)

    def parse_args(self, args=None):
        """Run ArgumentParser and filter out special arguments.

        There are some arguments considered "special" (see the SPECIAL_ARG_NAMES
        constant) and they aren't passed as parameters when the operation is
        run.

        """

        args_namespace = self.parser.parse_args(args)

        global_args = {
            k: v
            for k, v in vars(args_namespace).items()
            if k in self.SPECIAL_ARG_NAMES
        }

        operation_args = {
            k: v
            for k, v in vars(args_namespace).items()
            if k not in self.SPECIAL_ARG_NAMES
        }

        return global_args, operation_args



import random

import kaiju
from kaiju.arguments import Data, Float, Int, RawString, Bool

trilotest_database_name = "trilotest"

trilotest = kaiju.Kaiju(trilotest_database_name)
trilotest.connect_database()


@trilotest(output="bigdata")
def generate_uniform(a, b, sample_size, seed=1138):
    random.seed(seed)
    for _ in xrange(sample_size):
        yield random.randint(a, b)


@trilotest(sample=Data("bigdata"), output="bigdata")
def squared(sample):
    for x in sample:
        yield x * x


@trilotest(sample=Data("bigdata"))
def mean(sample):
    def sum_vec(x, y):
        return x[0] + y[0], x[1] + y[1]
    mean_data_gen = ((x, 1) for x in sample)
    sample_sum, sample_count = reduce(sum_vec, mean_data_gen, (0.0, 0.0))
    yield sample_sum / sample_count


@trilotest(my_int=Int(), my_float=Float(), my_string=RawString(), my_bool=Bool())
def some_arguments_testing(my_int, my_float, my_string="string", my_bool=False):
    yield dict(my_bool=my_bool, my_int=my_int, my_float=my_float, my_string=my_string)


@trilotest(pokemons=Data("pokemon"))
def provide_pokemons(pokemons):
    return pokemons



import re

from bson.json_util import loads
import bson.objectid  # For BSON's ObjectId.


class Argument(object):

    def __init__(self, **extra_config):
        self.config = dict(extra_config)
        if self.parse:
            self.config["type"] = self.parse

    def parse(self, string):
        return string

    def process(self, app, value):
        return value


class Var(Argument):

    def parse(self, string):
        try:
            return loads(string)
        except ValueError:
            raise ValueError(
                "Cannot deserialize argument: \"{}\". Maybe it is a string missing quotes?"
                .format(string)
            )


class File(Argument):

    def __init__(self, mode="r", **extra_config):
        super(File, self).__init__(**extra_config)
        self._mode = mode

    def process(self, app, value):
        return open(value, mode=self._mode) if value is not None else None


class DictsFile(Argument):
    """A "dicts file" is a text file that containts a JSON-coded
dictionary in each line. Similar to a MongoDB collection of documents,
each line from the file is a document.

    """

    def process(self, app, value):
        return (loads(x) for x in open(value))


class Data(Argument):

    def __init__(self, collection, mongo_cursor=False, **extra_config):
        super(Data, self).__init__(**extra_config)
        self._collection = collection
        self.collection = collection
        self.mongo_cursor = mongo_cursor

    def parse(self, value):
        return oid_parser(value)

    def process(self, app, value):
        return app.fetch(value, collection=self._collection, mongo_cursor=self.mongo_cursor)


class ObjectId(Argument):

    def __init__(self, **extra_config):
        super(ObjectId, self).__init__(**extra_config)

    def parse(self, value):
        return oid_parser(value)


class RawString(Argument):
    pass


class Int(Argument):

    def parse(self, value):
        return int(value)


class Float(Argument):

    def parse(self, value):
        return float(value)


class Bool(Var):

    def __init__(self, action="store_true", **extra_config):
        super(Bool, self).__init__(action=action, **extra_config)

    parse = None


def oid_parser(value):
    """
    Parse a `value` representing a MongoDB ObjectID.

    The parameter `value` can be a JSON string, e.g. `{'$oid': '...'}`, or a
    24-digit hexadecimal value.
    """
    m = re.match("[a-fA-F0-9]{24}", value)

    if m is not None:  # RegExp matched: it's a hexadecimal string.
        return bson.objectid.ObjectId(value)
    else:  # Check if it's a JSON string.
        msg = "Couldn't parse given string as an ObjectId."
        assert type(loads(value)) == bson.objectid.ObjectId, msg

        return loads(value)



import setuptools

setuptools.setup(
    name="kaiju",
    version="0.1.0",
    url="http://www.geekie.com.br",
    maintainer="Geekie",
    maintainer_email="geekie@geekie.com.br",
    description="Data analysis framework to log operations and interface with MongoDB.",
    packages=["kaiju"],
    include_package_data=True,
    setup_requires=["setuptools_git>=0.4.2"],
    install_requires=["pymongo"],
)
